// ZDriver_Be.cpp
/* Copyright (c) 1999 Andrew Green. ag@electricmagic.com
Permission is hereby granted, free of charge, to any person obtaining a copy of this software
and associated documentation files  (the "Software"), to deal in the Software without
restriction, including without limitation the rights to use, copy, modify, merge, publish,
distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the
Software is furnished to do so, subject to the following conditions:
The above copyright notice and this permission notice shall be included in all copies or
substantial portions of the Software.
THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING 
BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND 
NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, 
DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING 
FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE 
SOFTWARE.
*/

#include "ZDriver_Be.h"
#include <kernel/OS.h> // For port stuff
#include <stdlib.h> // For malloc/free
#include <stdarg.h>
#include <stdio.h>

// ==================================================
// Functions called by the driver infrastructure

int32 api_version = B_CUR_DRIVER_API_VERSION;

status_t init_hardware()
	{ return ZDriver_Be_Driver::sInitHardware(); }

status_t init_driver()
	{ return ZDriver_Be_Driver::sInitDriver(); }

void uninit_driver()
	{
	ZDriver_Be_Driver::sDriver->Finalize();
	delete ZDriver_Be_Driver::sDriver;
	}

const char** publish_devices()
	{ return ZDriver_Be_Driver::sDriver->PublishDeviceNames(); }

device_hooks* find_device(const char* inName)
	{ return &ZDriver_Be_Driver::sDeviceHooks; }

// ==================================================
// Implementations of necessary C++ runtime support routines. Note that we do not use
// exceptions in this code, although there are several places where it might make sense
// to do so. As they're not used, we don't need to provide runtime support.

void* __builtin_new (size_t inSize)
	{
	return ::malloc(inSize);
	}

void __builtin_delete(void* inPtr)
	{
	::free(inPtr);
	}

extern "C" void __pure_virtual()
	{
	}

// ====================================================================================================
#pragma mark -
#pragma mark ZDriver_Be_Driver

// Singleton ZDriver_Be_Driver instance
ZDriver_Be_Driver* ZDriver_Be_Driver::sDriver = nil;

device_hooks ZDriver_Be_Driver::sDeviceHooks =
	{ sOpen, sClose, sFree, sControl, sRead, sWrite, nil, nil, nil, nil };

// ==================================================
// You must provide implementations of these two methods in your own code. These are just examples.

#if 0
status_t ZDriver_Be_Driver::sInitHardware()
	{
	return B_OK;
	}

status_t ZDriver_Be_Driver::sInitDriver()
	{
	MyDriver* theDriver = new MyDriver;
	if (theDriver->Initialize())
		return B_OK;
	theDriver->Finalize();
	delete theDriver;
	return B_ERROR;
	}
#endif

// ==================================================

status_t ZDriver_Be_Driver::sOpen(const char* inName, uint32 inFlags, void** outCookie)
	{
	if (ZDriver_Be_Device* theDevice = sDriver->FindDeviceByName(inName))
		return theDevice->Open(inName, inFlags, *reinterpret_cast<ZDriver_Be_DeviceInstance**>(outCookie));
	return B_ERROR;
	}

status_t ZDriver_Be_Driver::sClose(void* inCookie)
	{ return reinterpret_cast<ZDriver_Be_DeviceInstance*>(inCookie)->Close(); }

status_t ZDriver_Be_Driver::sFree(void* inCookie)
	{ return reinterpret_cast<ZDriver_Be_DeviceInstance*>(inCookie)->Free(); }
	
status_t ZDriver_Be_Driver::sControl(void* inCookie, uint32 inOperation, void* inOutData, size_t inLen)
	{ return reinterpret_cast<ZDriver_Be_DeviceInstance*>(inCookie)->Control(inOperation, inOutData, inLen); }

status_t ZDriver_Be_Driver::sRead(void* inCookie, off_t inPos, void* outData, size_t* inOutLen)
	{ return reinterpret_cast<ZDriver_Be_DeviceInstance*>(inCookie)->Read(inPos, outData, *inOutLen); }

status_t ZDriver_Be_Driver::sWrite(void* inCookie, off_t inPos, const void* inData, size_t* inOutLen)
	{ return reinterpret_cast<ZDriver_Be_DeviceInstance*>(inCookie)->Write(inPos, inData, *inOutLen); }

// ==================================================

ZDriver_Be_Driver::ZDriver_Be_Driver()
	{
//	ZAssert(sDriver == nil);
	sDriver = this;

	for (size_t x = 0; x < kMaxCards; ++x)
		fCards[x] = nil;
	fCountCards = 0;
	fDeviceNamesCached = nil;
	}

ZDriver_Be_Driver::~ZDriver_Be_Driver()
	{
//	ZAssert(sDriver == this);
//	ZAssert(fCountCards == 0);
//	ZAssert(fDeviceNamesCached == nil);

	sDriver = nil;
	}

bool ZDriver_Be_Driver::Initialize()
	{
	return true;
	}

void ZDriver_Be_Driver::Finalize()
	{
	while (fCountCards)
		{
		fCards[fCountCards-1]->Finalize();
		delete fCards[fCountCards-1];
		fCards[fCountCards-1] = nil;
		--fCountCards;
		}

	if (fDeviceNamesCached)
		{
		::free(fDeviceNamesCached);
		fDeviceNamesCached = nil;
		}
	}


const char** ZDriver_Be_Driver::PublishDeviceNames()
	{
	if (!fDeviceNamesCached)
		{
		size_t totalNames = 0;
		for (size_t x = 0; x < fCountCards; ++x)
			totalNames += fCards[x]->CountNames();

		fDeviceNamesCached = reinterpret_cast<const char**>(::malloc((totalNames + 1) * sizeof(const char*)));

		size_t currentOffset = 0;
		for (size_t x = 0; x < fCountCards; ++x)
			{
			size_t namesInCard = fCards[x]->CountNames();
			for (size_t y = 0; y < namesInCard; ++y)
				{
				fDeviceNamesCached[currentOffset] = fCards[x]->GetNameByIndex(y);
				++currentOffset;
				}
			}
		fDeviceNamesCached[currentOffset] = nil;
		}

	return fDeviceNamesCached;
	}

ZDriver_Be_Device* ZDriver_Be_Driver::FindDeviceByName(const char* inName)
	{
	for (size_t x = 0; x < fCountCards; ++x)
		{
		if (ZDriver_Be_Device* theDevice = fCards[x]->FindDeviceByName(inName))
			return theDevice;
		}
	return nil;
	}

bool ZDriver_Be_Driver::AddCard(ZDriver_Be_Card* inCard)
	{
	if (inCard->Initialize())
		{
		fCards[fCountCards] = inCard;
		++fCountCards;
		return true;
		}
	inCard->Finalize();
	delete inCard;
	return false;
	}

// ====================================================================================================
#pragma mark -
#pragma mark ZDriver_Be_Card

ZDriver_Be_Card::ZDriver_Be_Card()
	{
	for (size_t x = 0; x < kMaxDevices; ++x)
		fDevices[x] = nil;
	fCountDevices = 0;
	}

ZDriver_Be_Card::~ZDriver_Be_Card()
	{
//	ZAssert(fCountDevices == 0);
	}

bool ZDriver_Be_Card::Initialize()
	{
	return true;
	}

void ZDriver_Be_Card::Finalize()
	{
	while (fCountDevices)
		{
		fDevices[fCountDevices-1]->Finalize();
		delete fDevices[fCountDevices-1];
		fDevices[fCountDevices-1] = nil;
		--fCountDevices;
		}
	}

size_t ZDriver_Be_Card::CountNames()
	{
	size_t count = 0;
	for (size_t x = 0; x < fCountDevices; ++x)
		count += fDevices[x]->CountNames();
	return count;
	}

const char* ZDriver_Be_Card::GetNameByIndex(size_t inNameIndex)
	{
	for (size_t x = 0; x < fCountDevices; ++x)
		{
		size_t currentCount = fDevices[x]->CountNames();
		if (inNameIndex < currentCount)
			{
			return fDevices[x]->GetNameByIndex(inNameIndex);
			}
		inNameIndex -= currentCount;
		}
	return nil;
	}

ZDriver_Be_Device* ZDriver_Be_Card::FindDeviceByName(const char* inName)
	{
	for (size_t x = 0; x < fCountDevices; ++x)
		{
		size_t namesInDevice = fDevices[x]->CountNames();
		for (size_t y = 0; y < namesInDevice; ++y)
			{
			if (!::strcmp(inName, fDevices[x]->GetNameByIndex(y)))
				return fDevices[x];
			}
		}
	return nil;
	}

bool ZDriver_Be_Card::AddDevice(ZDriver_Be_Device* inDevice)
	{
	if (inDevice->Initialize())
		{
		fDevices[fCountDevices] = inDevice;
		++fCountDevices;
		return true;
		}

	inDevice->Finalize();
	delete inDevice;
	}

// ====================================================================================================
#pragma mark -
#pragma mark ZDriver_Be_Device

ZDriver_Be_Device::ZDriver_Be_Device()
	{}

ZDriver_Be_Device::~ZDriver_Be_Device()
	{}

bool ZDriver_Be_Device::Initialize()
	{
	return true;
	}

void ZDriver_Be_Device::Finalize()
	{
	}

// ====================================================================================================
#pragma mark -
#pragma mark ZDriver_Be_DeviceInstance

ZDriver_Be_DeviceInstance::ZDriver_Be_DeviceInstance()
	{}

ZDriver_Be_DeviceInstance::~ZDriver_Be_DeviceInstance()
	{}

// ====================================================================================================
#pragma mark -
#pragma mark ZDebug

const char* ZDebug_FormatMessage(const char* message, ...)
	{
	static char debugMessageString[256];
	va_list args;
	va_start(args, message);
	short count = vsprintf(debugMessageString, message, args);
	va_end(args);
	return debugMessageString;
	}

#define message_AssertionAndUser "Assertion failed: %s. %s, %s:%d\n"
#define message_AssertionOnly "Assertion failed: %s, %s:%d\n"
#define message_UserOnly "%s, %s:%d\n"
#define message_None "%s:%d\n"

void ZDebug_DisplayMessage(int level, ZDebug_Action action, const char* filename, int line, const char* assertionMessage, const char* userMessage)
	{
	if (port_id thePortID = ::find_port("portlistener"))
		{
		char realDebugMessageString[256];
	
		if (assertionMessage)
			{
			if (userMessage)
				::sprintf(realDebugMessageString, message_AssertionAndUser, assertionMessage, userMessage, filename, line);
			else
				::sprintf(realDebugMessageString, message_AssertionOnly, assertionMessage, filename, line);
			}
		else
			{
			if (userMessage)
				::sprintf(realDebugMessageString, message_UserOnly, userMessage, filename, line);
			else
				::sprintf(realDebugMessageString, message_None, filename, line);
			}

		::write_port(thePortID, 0, realDebugMessageString, strlen(realDebugMessageString)+1);
		}

//	if (action == eDebugAction_Throw)
//		throw ZDebug_Exception(realDebugMessageString);
//
//	if (action == eDebugAction_Stop)
//		::debugger(realDebugMessageString);
//	else
//		{
//		::fprintf(stderr, realDebugMessageString);
//		::fprintf(stderr, "\n");
//		}
	}

